import re

from khmernltk.utils import constants


def correct_text(text: str):
    for token, normalized_token in constants.mistokenized_dict.items():
        text = text.replace(token, normalized_token)
    return text


def cleanup_str(text: str):
    text = text.strip(constants.SEPARATOR).strip()
    text = text.replace("  ", " ")  # clean up 2 spaces to 1
    text = text.replace(" ", "\u200b \u200b")  # ensure 200b around space
    # clean up
    text = text.replace("\u200b\u200b", "\u200b")  # clean up dupe 200b
    text = text.replace("\u200b\u200b", "\u200b")  # in case multiple
    text = correct_text(text)  # assume space has 200b wrapped around

    # remove special characters
    text = text.replace("\u2028", "")  # line separator
    text = text.replace("\u200a", "")  # hair space
    text = text.strip().replace("\n", "").replace("  ", " ")
    return text


def post_process(text, separator):
    text = text.strip(separator)
    return re.sub(f"(?:{separator}| )+", separator, text)


def is_khmer_char(ch: str):
    if (ch >= "\u1780") and (ch <= "\u17ff"):
        return True
    if ch in constants.KHSYM:
        return True
    if ch in constants.KHLUNAR:
        return True
    return False


def is_start_of_kcc(ch: str):
    if is_khmer_char(ch):
        if ch in constants.KHCONST:
            return True
        if ch in constants.KHSYM:
            return True
        if ch in constants.KHNUMBER:
            return True
        if ch in constants.KHLUNAR:
            return True
        return False
    return True


# kcc base - must surround space with \u200b using cleanupstr()


def seg_kcc(str_sentence: str):
    segs = []
    cur = ""
    sentence = str_sentence
    for word in sentence.split(constants.SEPARATOR):
        for i, c in enumerate(word):
            cur += c
            nextchar = word[i + 1] if (i + 1 < len(word)) else ""

            # cluster non-khmer chars together
            if not is_khmer_char(c) and nextchar != " " and nextchar != "" and not is_khmer_char(nextchar):
                continue
            # cluster number together
            if c in constants.KHNUMBER and nextchar in constants.KHNUMBER:
                continue

            # cluster non-khmer together
            # non-khmer character has no cluster
            if not is_khmer_char(c) or nextchar == " " or nextchar == "":
                segs.append(cur)
                cur = ""
            elif is_start_of_kcc(nextchar) and not (c in constants.KHSUB):
                segs.append(cur)
                cur = ""
        # add space back after split
        # segs.append(" ")
    return segs  # [:-1] # trim last space



import os
from typing import List, Union

from khmernltk.utils import constants
from khmernltk.utils.data import cleanup_str, seg_kcc
from khmernltk.utils.file_utils import load_model
from khmernltk.word_tokenize.features import create_kcc_features

# sklearn_crf_ner_alt_0.9725.sav / sklearn_crf_ner_10000.sav
model_path = os.path.join(os.path.dirname(__file__), "sklearn_crf_ner_10000.sav")
crf_model = None


def word_tokenize(text: str, separator: str = "-", return_tokens: bool = True) -> Union[List, str]:
    """Khmer language word tokenization

    Args:
        text(str): Raw text
        separator(str, optional): Token - separator in case return_tokens = True. Defaults to "-".
        return_tokens(bool, optional): Whether return a tokenized text or a list of tokens. Defaults to True.

    Returns:
        Union[List, str]: Tokens or tokenized text, separated by the separator
    """
    global crf_model
    if crf_model is None:
        crf_model = load_model(model_path)

    text = cleanup_str(text)
    skcc = seg_kcc(text)

    features = create_kcc_features(skcc)
    pred = crf_model.predict([features])

    tkcc = []
    for k in features:
        tkcc.append(k["kcc"])
    complete = ""
    tokens = []
    for i, p in enumerate(pred[0]):
        if p == "1" or i == 0:
            tokens.append(tkcc[i])
        else:
            tokens[-1] += tkcc[i]
    if return_tokens:
        return tokens

    complete = separator.join(tokens)
    complete = complete.replace(separator + " " + separator, " ")

    return complete



import pytest

from khmernltk import pos_tag, sentence_tokenize, word_tokenize


def test_pos_tag():
    text = "ខួបឆ្នាំទី២៨! ២៣ តុលា ស្មារតីផ្សះផ្សាជាតិរវាងខ្មែរនិងខ្មែរ ឈានទៅបញ្ចប់សង្រ្គាម នាំពន្លឺសន្តិភាព និងការរួបរួមជាថ្មី"
    return pos_tag(text)


def test_sentence_tokenize():
    text = "ខួបឆ្នាំទី២៨! ២៣ តុលា ស្មារតីផ្សះផ្សាជាតិរវាងខ្មែរនិងខ្មែរ ឈានទៅបញ្ចប់សង្រ្គាម នាំពន្លឺសន្តិភាព និងការរួបរួមជាថ្មី"
    return sentence_tokenize(text)


def test_word_tokenize():
    text = "ខួបឆ្នាំទី២៨! ២៣ តុលា ស្មារតីផ្សះផ្សាជាតិរវាងខ្មែរនិងខ្មែរ ឈានទៅបញ្ចប់សង្រ្គាម នាំពន្លឺសន្តិភាព និងការរួបរួមជាថ្មី"
    return word_tokenize(text)



import glob
import os
import pickle
import re
import shutil

from tqdm import tqdm

from khmernltk.utils import constants
from khmernltk.utils.data import cleanup_str
from khmernltk.utils.log_utils import logger


def get_data_from_folder(data_dir: str, max_sentences=100000):
    path = data_dir + "/*_seg_200b.txt"  # earlier format: *_seg.txt
    files = glob.glob(path)

    # global variables that use through out
    sentences = []
    seg_text = []
    orig_text = []
    # unique id of the article that can be matched to docId in meta.txt
    doc_ids = []
    for file in tqdm(files, desc=f"Getting data from {data_dir}", leave=True):
        filenum = re.search(r"\d+_", file).group(0)
        doc_ids.append(filenum.replace("_", ""))
        f = open(file, "r")
        lines = f.readlines()
        f.close()

        seg_text.append(lines)

        if len(seg_text) >= max_sentences:
            break

        # read orig text -- comment out (10K docs which do not have orig text)
        f = open(file.replace("_seg_200b.txt", "_orig.txt"), "r")
        lines = f.readlines()
        f.close()
        orig_text.append(lines)

    for text in seg_text:
        for sentence in text:
            sentences.append(cleanup_str(sentence))

    logger.info(f"number of file: {len(files)}")
    logger.info(f"number of doc_ids: {len(doc_ids)}")
    logger.info(f"number of sentences: {len(sentences)}")
    return sentences  # , orig_text


def _normalize_token(token):
    assert token.count("[") == token.count("]"), f"token {token} is not valid"
    token = token.replace("-", "")
    assert token[0] == token[-1], f"confused in specify token {token}"
    return token[0]


def cache_nova_text(tok_fp, tag_fp, output_dir):
    with open(tag_fp, "r", encoding="utf8") as f:
        tagged_sentences_text = f.read().strip().split("\n")
    with open(tok_fp, "r", encoding="utf8") as f:
        sentences_text = f.read().strip().split("\n")

    tagged_sentences = {index: tagged for index, tagged in map(lambda text: text.split("\t"), tagged_sentences_text)}
    sentences = {index: tagged for index, tagged in map(lambda text: text.split("\t"), sentences_text)}

    new_tagged_sentences_text = dict()
    new_sentences_text = dict()

    for sentence_id in tqdm(tagged_sentences.keys(), desc="Processing data"):
        new_sentence = ""
        new_tokens = []
        start_of_token = False
        for i, token in enumerate(tagged_sentences[sentence_id].split()):
            if "[" in token:
                new_sentence += constants.SEPARATOR + sentences[sentence_id].split()[i]
                #             token = _normalize_token(token)
                new_tokens.append(token)
                start_of_token = True
                continue

            if start_of_token:
                new_sentence += sentences[sentence_id].split()[i]
                new_tokens[-1] += "+" + token
            else:
                new_sentence += constants.SEPARATOR + sentences[sentence_id].split()[i]
                token = _normalize_token(token)
                new_tokens.append(token)

            if "]" in token:
                start_of_token = False
                new_tokens[-1] = _normalize_token(new_tokens[-1])

        new_sentence = new_sentence.strip(constants.SEPARATOR).strip()
        new_tagged_sentences_text[sentence_id] = " ".join(new_tokens).strip()
        new_sentences_text[sentence_id] = new_sentence
        assert len(new_tokens) == len(new_sentence.split(constants.SEPARATOR)), f"{sentence_id}"

    # new_tagged_sentences_text = OrderedDict(sorted(new_tagged_sentences_text.items()))
    # new_sentences_text = OrderedDict(sorted(new_sentences_text.items()))

    # recreate the output directory (cause we're writing in adding mode)
    if os.path.exists(output_dir):
        shutil.rmtree(output_dir)
    os.makedirs(output_dir)

    for sentence_id in tqdm(sorted(new_tagged_sentences_text.keys()), f"Saving data to {output_dir}"):
        id_ = sentence_id.split(".")
        doc_id = ".".join(id_[:-1])

        tok_fp = os.path.join(output_dir, doc_id + "_seg_200b.txt")
        with open(tok_fp, "a") as f:
            # f.write(sentence_id)
            # f.write("\t")
            f.write(new_sentences_text[sentence_id])
            f.write("\n")

        orig_fp = os.path.join(output_dir, doc_id + "_orig.txt")
        with open(orig_fp, "a") as f:
            # f.write(sentence_id)
            # f.write("\t")
            f.write(new_sentences_text[sentence_id].replace(constants.SEPARATOR, ""))
            f.write("\n")

        tag_fp = os.path.join(output_dir, doc_id + "_tag.txt")

        with open(tag_fp, "a") as f:
            # f.write(sentence_id)
            # f.write("\t")
            f.write(new_tagged_sentences_text[sentence_id])
            f.write("\n")


def save_model(model, fn, output_dir="../../outputs/"):
    os.makedirs(output_dir, exist_ok=True)
    fp = os.path.join(output_dir, fn)
    with open(fp, "wb") as f:
        pickle.dump(model, f)
    logger.info(f"Saved model to {fp}")


def load_model(fp):
    with open(fp, "rb") as f:
        model = pickle.load(f)
    logger.info(f"Loaded model from {fp}")
    return model



import pytest

from khmernltk import pos_tag, sentence_tokenize, word_tokenize

text = (
    "ខួបឆ្នាំទី២៨! ២៣ តុលា ស្មារតីផ្សះផ្សាជាតិរវាងខ្មែរនិងខ្មែរ ឈានទៅបញ្ចប់សង្រ្គាម នាំពន្លឺសន្តិភាព និងការរួបរួមជាថ្មី"
)
correct = "​ខួប​ឆ្នាំ​ទី ២៨ !​ ២៣ ​តុលា ​ស្មារតី​ផ្សះផ្សា​ជាតិ​រវាង​ខ្មែរ​និង​ខ្មែរ ​ឈាន​ទៅ​បញ្ចប់​សង្រ្គាម ​នាំ​ពន្លឺ​សន្តិភាព​ ​និង​ការរួបរួម​ជាថ្មី​"

# test sentence tokenizer
print("sentence-level tokenized text")
print(sentence_tokenize(text))

# test word tokenizer
sep = "_"
print("word-level tokenized text")
print(word_tokenize(text, separator=sep, return_tokens=True))
print("correct text")
print(correct.strip("\u200b").replace("\u200b", sep))

# test pos tagger
print(pos_tag(text))



import os
from typing import List, Tuple

from khmernltk import word_tokenize
from khmernltk.pos_tag.features import create_word_features
from khmernltk.utils.data import cleanup_str
from khmernltk.utils.file_utils import load_model

# sklearn_crf_pos_alt_0.9849.sav
model_path = os.path.join(os.path.dirname(__file__), "sklearn_crf_pos_alt_0.9846.sav")
crf_model = None


def pos_tag(text: str) -> List[Tuple[str, str]]:
    """Khmer language Part of speech tagging

    Args:
        text (str): Raw text

    Returns:
        List[Tuple[str, str]]: List of tuple (token, pos tag)
    """
    global crf_model
    if crf_model is None:
        crf_model = load_model(model_path)

    text = cleanup_str(text)
    tokens = word_tokenize.word_tokenize(text, return_tokens=True)
    features = create_word_features(tokens)
    pred = crf_model.predict([features])[0]
    return list(zip(tokens, pred))



from khmernltk.utils.data import seg_kcc


def word_to_features(sent, i):
    word = sent[i]
    # split word into Khmer character clusters
    kccs = seg_kcc(word)

    features = {
        "bias": 1.0,
        "len(kccs)": len(kccs),
        "word.isdigit()": word.isdigit(),
        "word": word,
    }

    for k in range(len(kccs)):
        features.update({f"kccs[{k}]": kccs[k]})

    if i > 0:
        word1 = sent[i - 1]
        kccs1 = seg_kcc(word1)
        features.update(
            {
                "-1:word.isdigit()": word1.isdigit(),
                "-1:word": word1,
                "-1len(kccs)": len(kccs1),
            }
        )
        for k in range(len(kccs1)):
            features.update({f"-1kccs1[{k}]": kccs1[k]})
    else:
        features["BOS"] = True

    if i < len(sent) - 1:
        word1 = sent[i + 1]
        kccs1 = seg_kcc(word1)
        features.update(
            {
                "+1:word.isdigit()": word1.isdigit(),
                "+1:word": word1,
                "+1len(kccs)": len(kccs1),
            }
        )
        for k in range(len(kccs1)):
            features.update({f"+1kccs1[{k}]": kccs1[k]})
    else:
        features["EOS"] = True

    return features


def create_word_features(sent):
    return [word_to_features(sent, i) for i in range(len(sent))]



import re
from typing import List

from khmernltk.sentence_tokenize.features import SENTENCE_SEPARATOR


def sentence_tokenize(text: str) -> List[str]:
    """Khmer language sentence tokenization

    Args:
        text (str): Raw text

    Returns:
        List[str]: List of sentences
    """
    sentences = re.split("(?<=[" + "".join(SENTENCE_SEPARATOR) + "])\s*", text)
    if sentences[-1]:
        return sentences
    return sentences[:-1]



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

requirements = ["sklearn-crfsuite", "tqdm"]

extras_requirements = {
    "test": [
        "pytest",
        "coverage",
    ],
}

setuptools.setup(
    name="khmer-nltk",
    version="1.6",
    description="A Khmer language processing toolkit",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/VietHoang1710/khmer-nltk",
    author="Phan Viet Hoang",
    author_email="phanviethoang1512@gmail.com",
    license="Apache License 2.0",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Natural Language :: English",
    ],
    python_requires=">3.5",
    packages=setuptools.find_packages(exclude=["data", "bin", "samples", "scripts"]),
    package_dir={"khmernltk": "khmernltk"},
    package_data={
        "khmernltk": [
            "pos_tag/sklearn_crf_pos_alt_0.9846.sav",
            "word_tokenize/sklearn_crf_ner_10000.sav",
            #   'word_tokenize/sklearn_crf_ner_alt_0.9725.sav'
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    extras_require=extras_requirements,
)



import logging
import logging.handlers
import re


class ColorCodes:
    grey = "\x1b[38;21m"
    green = "\x1b[1;32m"
    yellow = "\x1b[33;21m"
    red = "\x1b[31;21m"
    bold_red = "\x1b[31;1m"
    blue = "\x1b[1;34m"
    light_blue = "\x1b[1;36m"
    purple = "\x1b[1;35m"
    reset = "\x1b[0m"


class ColorizedArgsFormatter(logging.Formatter):
    arg_colors = [ColorCodes.purple, ColorCodes.light_blue]
    level_fields = ["levelname", "levelno"]
    level_to_color = {
        logging.DEBUG: ColorCodes.grey,
        logging.INFO: ColorCodes.green,
        logging.WARNING: ColorCodes.yellow,
        logging.ERROR: ColorCodes.red,
        logging.CRITICAL: ColorCodes.bold_red,
    }

    def __init__(self, fmt: str):
        super().__init__()
        self.level_to_formatter = {}

        def add_color_format(level: int):
            color = ColorizedArgsFormatter.level_to_color[level]
            _format = fmt
            for fld in ColorizedArgsFormatter.level_fields:
                search = "(%\(" + fld + "\).*?s)"
                _format = re.sub(search, f"{color}\\1{ColorCodes.reset}", _format)
            formatter = logging.Formatter(_format)
            self.level_to_formatter[level] = formatter

        add_color_format(logging.DEBUG)
        add_color_format(logging.INFO)
        add_color_format(logging.WARNING)
        add_color_format(logging.ERROR)
        add_color_format(logging.CRITICAL)

    @staticmethod
    def rewrite_record(record: logging.LogRecord):
        if not BraceFormatStyleFormatter.is_brace_format_style(record):
            return

        msg = record.msg
        msg = msg.replace("{", "_{{")
        msg = msg.replace("}", "_}}")
        placeholder_count = 0
        # add ANSI escape code for next alternating color before each formatting parameter
        # and reset color after it.
        while True:
            if "_{{" not in msg:
                break
            color_index = placeholder_count % len(ColorizedArgsFormatter.arg_colors)
            color = ColorizedArgsFormatter.arg_colors[color_index]
            msg = msg.replace("_{{", color + "{", 1)
            msg = msg.replace("_}}", "}" + ColorCodes.reset, 1)
            placeholder_count += 1

        record.msg = msg.format(*record.args)
        record.args = []

    def format(self, record):
        orig_msg = record.msg
        orig_args = record.args
        formatter = self.level_to_formatter.get(record.levelno)
        self.rewrite_record(record)
        formatted = formatter.format(record)

        # restore log record to original state for other handlers
        record.msg = orig_msg
        record.args = orig_args
        return formatted


class BraceFormatStyleFormatter(logging.Formatter):
    def __init__(self, fmt: str):
        super().__init__()
        self.formatter = logging.Formatter(fmt)

    @staticmethod
    def is_brace_format_style(record: logging.LogRecord):
        if len(record.args) == 0:
            return False

        msg = record.msg
        if "%" in msg:
            return False

        count_of_start_param = msg.count("{")
        count_of_end_param = msg.count("}")

        if count_of_start_param != count_of_end_param:
            return False

        if count_of_start_param != len(record.args):
            return False

        return True

    @staticmethod
    def rewrite_record(record: logging.LogRecord):
        if not BraceFormatStyleFormatter.is_brace_format_style(record):
            return

        record.msg = record.msg.format(*record.args)
        record.args = []

    def format(self, record):
        orig_msg = record.msg
        orig_args = record.args
        self.rewrite_record(record)
        formatted = self.formatter.format(record)

        # restore log record to original state for other handlers
        record.msg = orig_msg
        record.args = orig_args
        return formatted


def custom_logger(name="khmer-nltk"):
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.DEBUG)
    console_format = "| %(asctime)s | %(levelname)s | %(name)s | %(message)s |"

    colored_formatter = ColorizedArgsFormatter(console_format)
    console_handler.setFormatter(colored_formatter)
    logger.addHandler(console_handler)
    return logger


logger = custom_logger()



from khmernltk.utils import constants


def get_type(char: str):
    if char.lower() in constants.EN:
        return constants.NS
    if char in constants.KHCONST:
        return "C"
    if char in constants.KHVOWEL:
        return "W"
    if char in constants.KHNUMBER:
        return constants.NS
    if char in constants.KHSUB:
        return "S"
    if char in constants.KHDIAC:
        return "D"
    return constants.NS


# non-khmer character that we should not separate like number
# multiple characters are false


def is_no_space(k):
    if get_type(k[0]) == constants.NS:
        return True
    return False


def kcc_type(k):
    if len(k) == 1:
        return get_type(k)
    else:
        return "K" + str(len(k))


def kcc_to_features(kccs, i):
    maxi = len(kccs)
    kcc = kccs[i]

    features = {"kcc": kcc, "t": kcc_type(kcc), "ns": is_no_space(kcc)}
    if i >= 1:
        features.update(
            {
                "kcc[-1]": kccs[i - 1],
                "kcc[-1]t": kcc_type(kccs[i - 1]),
                "kcc[-1:0]": kccs[i - 1] + kccs[i],
                "ns-1": is_no_space(kccs[i - 1]),
            }
        )
    else:
        features["BOS"] = True

    if i >= 2:
        features.update(
            {
                "kcc[-2]": kccs[i - 2],
                "kcc[-2]t": kcc_type(kccs[i - 2]),
                "kcc[-2:-1]": kccs[i - 2] + kccs[i - 1],
                "kcc[-2:0]": kccs[i - 2] + kccs[i - 1] + kccs[i],
            }
        )
    if i >= 3:
        features.update(
            {
                "kcc[-3]": kccs[i - 3],
                "kcc[-3]t": kcc_type(kccs[i - 3]),
                "kcc[-3:0]": kccs[i - 3] + kccs[i - 2] + kccs[i - 1] + kccs[i],
                "kcc[-3:-1]": kccs[i - 3] + kccs[i - 2] + kccs[i - 1],
                "kcc[-3:-2]": kccs[i - 3] + kccs[i - 2],
            }
        )

    if i < maxi - 1:
        features.update(
            {
                "kcc[+1]": kccs[i + 1],
                "kcc[+1]t": kcc_type(kccs[i + 1]),
                "kcc[+1:0]": kccs[i] + kccs[i + 1],
                "ns+1": is_no_space(kccs[i + 1]),
            }
        )
    else:
        features["EOS"] = True

    if i < maxi - 2:
        features.update(
            {
                "kcc[+2]": kccs[i + 2],
                "kcc[+2]t": kcc_type(kccs[i + 2]),
                "kcc[+1:+2]": kccs[i + 1] + kccs[i + 2],
                "kcc[0:+2]": kccs[i + 0] + kccs[i + 1] + kccs[i + 2],
                "ns+2": is_no_space(kccs[i + 2]),
            }
        )
    if i < maxi - 3:
        features.update(
            {
                "kcc[+3]": kccs[i + 3],
                "kcc[+3]t": kcc_type(kccs[i + 3]),
                "kcc[+2:+3]": kccs[i + 2] + kccs[i + 3],
                "kcc[+1:+3]": kccs[i + 1] + kccs[i + 2] + kccs[i + 3],
                "kcc[0:+3]": kccs[i + 0] + kccs[i + 1] + kccs[i + 2] + kccs[i + 3],
            }
        )

    return features


def create_kcc_features(kccs):
    return [kcc_to_features(kccs, i) for i in range(len(kccs))]
